#include "ConvDialog.h"
#include "xom2.h"

#include <QApplication>
#include <QMessageBox>
#include <QSettings>
#include <QDir>
#include <QDirIterator>
#include <QTextStream>
#include <QTextEdit>
#include <QFileDialog>

int main(int argc, char *argv[])
{
	QApplication app(argc, argv);

	ConvDialog *dialog = new ConvDialog(argc, argv);
	if(dialog->gamepath=="NONE") return 0;
	
	dialog->show();
		
	return app.exec();
}

bool xom2_renameDuplicateProperties(QDomElement& el)
{
    QDomNode n = el.firstChild();
    QString prevTag;
    int nTag=0;
    while(!n.isNull())
    {
        QDomElement e = n.toElement();
        if(!e.isNull())
        {
            QString tag = e.tagName();
            if(tag==prevTag) e.setTagName(tag+"("+QString::number(++nTag)+")");
            else nTag = 0;
            prevTag = tag;
        }
        n = n.nextSibling();
    }
    return false;
}

QString xom2_toProperty(QDomElement& e)
{
    if(!e.isNull())
    {
        if(e.hasAttribute("r")&&e.hasAttribute("g")&&e.hasAttribute("b")&&e.hasAttribute("a"))
        {
            return "@color "+e.attribute("r")+" "+e.attribute("g")+" "+e.attribute("b")+" "+e.attribute("a");
        }
        if(e.hasAttribute("r")&&e.hasAttribute("g")&&e.hasAttribute("b"))
        {
            return "@color "+e.attribute("r")+" "+e.attribute("g")+" "+e.attribute("b");
        }
        if(e.hasAttribute("x")&&e.hasAttribute("y")&&e.hasAttribute("z"))
        {
            return "@vector "+e.attribute("x")+" "+e.attribute("y")+" "+e.attribute("z");
        }
        if(e.hasAttribute("x")&&e.hasAttribute("y"))
        {
            return "@vector "+e.attribute("x")+" "+e.attribute("y");
        }
        if(e.hasAttribute("href"))
        {
            return "@href "+e.attribute("href");
        }
        return e.text();
    }
    return QString();
}

void resource_set(QString& data, QString type, const QString& name, const QString& value)
{
    if     (type=="XIntResourceDetails")    type="#int";
    else if(type=="XUintResourceDetails")   type="#uint";
    else if(type=="XFloatResourceDetails")  type="#float";
    else if(type=="XStringResourceDetails") type="#string";
    else if(type=="XVectorResourceDetails") type="#vector";
    else if(type=="XColorResourceDetails")  type="#color";
    
    QString newdata;
    QTextStream ts(&data);
    QString s;
    do
    {
        s = ts.readLine();
        if(!s.isEmpty()) newdata = newdata+s+"\n";
    } while(s != type && !s.isEmpty());
    if(s.isEmpty())
    {
        newdata = newdata+type+"\n"+name+"="+value;
        data = newdata;
        return;
    }
    else
    {
        bool done = false;
        do
        {
            s = ts.readLine();
            if(s.isEmpty()) break;
            if(s.section('=',0,0)==name)
            {
                newdata = newdata+s.section('=',0,0)+"="+value+"\n";
                s = ts.readLine();
                done = true;
                break;
            }
            if(s.startsWith("#")) break;
            newdata = newdata+s+"\n";
        } while(!s.isEmpty());
        if(!done & (s.isEmpty() || s.startsWith("#")))
        {
            newdata = newdata+name+"="+value+"\n";
        }
        while(!s.isEmpty())
        {
            newdata = newdata+s+"\n";
            s = ts.readLine();
        }
        data = newdata;
        data.chop(1);
    }
}

///////////////////////////////////////////////////////////////////////////////////////
// Constructor
// Sets the interface up, etc...

ConvDialog::ConvDialog(int argc, char* argv[]) : QDialog()
{
    // Retrieve the game install path from the registry
    QSettings reg("HKEY_LOCAL_MACHINE\\SOFTWARE\\CODEMASTERS\\Worms 4 Mayhem", QSettings::NativeFormat);
    gamepath = reg.value("InstallDir", "NONE").toString();
    if(gamepath == "NONE") // The key doesn't exist, game is probably not installed. Generate error message, and quit.
    {
        QMessageBox::critical(this, "Error !", "Could not find Worms 4: Mayhem directory. Please reinstall the game.");
        return;
    }
    lastdir1 = gamepath;
    lastdir2 = gamepath;
    
    ui.setupUi(this);
    
	// Connect the signals and slots
	connect(ui.quitButton, SIGNAL(clicked()), qApp, SLOT(quit()));
	connect(ui.convertButton, SIGNAL(clicked()), this, SLOT(doConvert()));
	connect(ui.fileButton, SIGNAL(clicked()), this, SLOT(selectSource()));
	connect(ui.targetButton, SIGNAL(clicked()), this, SLOT(selectDestination()));
}

///////////////////////////////////////////////////////////////////////////////////////
// Destructor. Do nothing for now...

ConvDialog::~ConvDialog()
{
    
}


void ConvDialog::selectSource()
{
    QString file = QFileDialog::getOpenFileName(this, "Select XML file to convert", lastdir1, "XML Tweak File (*.xml)");
    if(!file.isNull())
    {
        ui.fileEdit->setText(file);
        lastdir1 = QFileInfo(file).absolutePath();
    }
}

void ConvDialog::selectDestination()
{
    QString dir = QFileDialog::getExistingDirectory(this, "Select directory for converted files", lastdir2);
    if(!dir.isNull())
    {
        ui.targetEdit->setText(dir);
        lastdir2 = dir;
    }
}

void ConvDialog::doConvert()
{
    QFile file(ui.fileEdit->text());
    QDir dir(ui.targetEdit->text());
    if(!file.exists()) {QMessageBox::critical(this, "Error !", "File "+ui.fileEdit->text()+" does not exist !");return;}
    QString refname = QFileInfo(ui.fileEdit->text()).fileName();
    QFile ref(gamepath + "/data/Tweak/" + refname);
    if(!ref.exists())  {QMessageBox::critical(this, "Error !", "Could not find reference "+refname+" in data\\Tweak !");return;}
    
    QDomDocument doc, refdoc;
    QDomElement xom, refxom;
    
    if(!xom2_load(doc, xom, ui.fileEdit->text())) {QMessageBox::critical(this, "Error !", "Unable to load XOM data from "+ui.fileEdit->text());return;}
    if(!xom2_load(refdoc, refxom, gamepath + "/data/Tweak/" + refname)) {QMessageBox::critical(this, "Error !", "Unable to load XOM data from "+gamepath + "/data/Tweak/" + refname);return;}
    
    refname.chop(4); // Remove the extension from the tweak file
    dir.mkpath("Tweak");
    dir.cd("Tweak"); // Make all the necessary directories if needed
    dir.mkdir(refname);
    dir.cd(refname);
    
    // Already a resource file ? Load its contents.
    QString restwk;
    QFile res(dir.path()+"/resource.twk");
    // end
    
    
    // Iterate through the objects in the tweak
    QDomNode n = xom.firstChild();
    while(!n.isNull())
    {
        QDomElement obj = n.toElement();
        if(!obj.isNull())
        {
            if(obj.tagName()=="XDataBank") {} // Skip the databank
            else if(obj.tagName()=="SKIP") {} // Skip objects whose container has already been read
            else if(obj.tagName()=="XContainerResourceDetails") // Found a container, try to see if it exists in the original tweak
            {
                QDomElement refobj=xom2_getObject(refxom, obj.attribute("id"));
                if(!refobj.isNull()) // Found it
                {
                    QString data;
                    QTextStream ts(&data);
                    
                    QString prop,prop0;
                    prop0 = xom2_getProperty(refobj, "Flags");
                    prop  = xom2_getProperty(obj   , "Flags");
                    if(prop!=prop0) ts << "#flags " << prop << "\n";
                    QString href = xom2_getProperty(obj, "Value");
                    href = href.mid(6);
                    QDomElement obj2 = xom2_getObject(xom, href);
                    obj2.setTagName("SKIP"); // Set the tag name to SKIP so this object won't be processed later
                    // Iterate through its properties
                    QDomNode n2 = obj2.firstChild();
                    QDomElement refobj2 = xom2_getObject(refxom, href);
                    
                    xom2_renameDuplicateProperties(obj2);
                    xom2_renameDuplicateProperties(refobj2);
                    while(!n2.isNull())
                    {
                        QDomElement pr = n2.toElement();
                        if(!pr.isNull())
                        {
                            QString pr2 = xom2_getProperty(refobj2, pr.tagName());
                            if(pr2!=xom2_toProperty(pr))
                            {
                                ts << pr.tagName() << "=" << xom2_toProperty(pr) << "\n";
                            }
                        }
                        n2 = n2.nextSibling();
                    }
                    
                    // If there have been modifications (different properties) then write file
                    if(!data.isEmpty())
                    {
                        QFile otwk(dir.path()+"/"+obj.attribute("id")+".twk");
                        otwk.open(QIODevice::WriteOnly);
                        otwk.write(data.toAscii());
                        otwk.close();
                    }
                }
                else // Not found, create it then
                {
                    QString data;
                    QTextStream ts(&data);
                    
                    QString prop;
                    prop  = xom2_getProperty(obj, "Flags");
                    if(prop!="80") ts << "#flags " << prop << "\n";
                    QString href = xom2_getProperty(obj, "Value");
                    href = href.mid(6);
                    QDomElement obj2 = xom2_getObject(xom, href);
                    
                    ts << "#class " << obj2.tagName() << "\n";
                    obj2.setTagName("SKIP"); // Set the tag name to SKIP so this object won't be processed later
                    
                    // Iterate through its properties
                    QDomNode n2 = obj2.firstChild();
                    
                    xom2_renameDuplicateProperties(obj2);
                    while(!n2.isNull())
                    {
                        QDomElement pr = n2.toElement();
                        if(!pr.isNull())
                        {
                            ts << pr.tagName() << "=" << xom2_toProperty(pr) << "\n";
                        }
                        n2 = n2.nextSibling();
                    }
                    
                    QFile otwk(dir.path()+"/"+obj.attribute("id")+".twk");
                    otwk.open(QIODevice::WriteOnly);
                    otwk.write(data.toAscii());
                    otwk.close();
                }
            }
            else if(obj.tagName()=="XIntResourceDetails"    | obj.tagName()=="XUintResourceDetails"   | obj.tagName()=="XFloatResourceDetails"
                  | obj.tagName()=="XStringResourceDetails" | obj.tagName()=="XVectorResourceDetails" | obj.tagName()=="XColorResourceDetails")
            {
                // Does the resource object exist in the original tweak ?
                QDomElement refobj=xom2_getObject(refxom, obj.attribute("id"));
                if(refobj.isNull()) // No, just create it and add it to the resource file then
                {
                    resource_set(restwk, obj.tagName(), obj.attribute("id"), xom2_getProperty(obj, "Value"));
                }
                else // Yes, check its value
                {
                    if(xom2_getProperty(obj, "Value")!=xom2_getProperty(refobj, "Value"))
                    {
                        // Not the same value as the original object, add it to the resource file
                        resource_set(restwk, obj.tagName(), obj.attribute("id"), xom2_getProperty(obj, "Value"));
                    }
                }
            }
            else // The object is not a resource object, and it doesn't have a container, check if it exists in the original tweak
            {
                QDomElement refobj=xom2_getObject(refxom, obj.attribute("id"));
                if(!refobj.isNull()) // Found it
                {
                    QString data;
                    QTextStream ts(&data);
                    
                    // Iterate through its properties
                    QDomNode n2 = obj.firstChild();
                    
                    xom2_renameDuplicateProperties(obj);
                    xom2_renameDuplicateProperties(refobj);
                    while(!n2.isNull())
                    {
                        QDomElement pr = n2.toElement();
                        if(!pr.isNull())
                        {
                            QString pr2 = xom2_getProperty(refobj, pr.tagName());
                            if(pr2!=xom2_toProperty(pr))
                            {
                                ts << pr.tagName() << "=" << xom2_toProperty(pr) << "\n";
                            }
                        }
                        n2 = n2.nextSibling();
                    }
                    
                    // If there have been modifications (different properties) then write file
                    QFile otwk(dir.path()+"/"+obj.attribute("id")+".twk");
                    otwk.open(QIODevice::WriteOnly);
                    otwk.write(data.toAscii());
                    otwk.close();
                }
                else // Not found, create it then
                {
                    QString data;
                    QTextStream ts(&data);
                    
                    ts << "#nodb\n#class " << obj.tagName() << "\n";
                    
                    // Iterate through its properties
                    QDomNode n2 = obj.firstChild();
                    
                    xom2_renameDuplicateProperties(obj);
                    while(!n2.isNull())
                    {
                        QDomElement pr = n2.toElement();
                        if(!pr.isNull())
                        {
                            ts << pr.tagName() << "=" << xom2_toProperty(pr) << "\n";
                        }
                        n2 = n2.nextSibling();
                    }
                    
                    QFile otwk(dir.path()+"/"+obj.attribute("id")+".twk");
                    otwk.open(QIODevice::WriteOnly);
                    otwk.write(data.toAscii());
                    otwk.close();
                }
            }
        }
        n = n.nextSibling();
    }
    
    if(!restwk.isEmpty())
    {
        res.open(QIODevice::WriteOnly|QIODevice::Text);
        res.write(restwk.toAscii());
        res.close();
    }
    QMessageBox::information(this, "xml2twk", "Done !");
}
